#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "image.h"

// Scaling images is known to be bugged, don't use it, okay?
 
void ExpandRow(JSAMPLE *start, JSAMPLE *source)
 {
  JSAMPLE *end = start + (source - start) * 3;
  
  for(; source != start; --source, end -= 3)
   end[-1] = end[-2] = end[-3] = source[-1];
 }
 
void CondenseRow(JSAMPLE *start, JSAMPLE *source)
 {
  JSAMPLE *end = start;
  
  for(; start != source; ++start, end += 3)
   *start = (end[0] + end[1] + end[2])/3;
 }
 
int isGreyPixel(JSAMPLE *data)
 {
  return abs(data[0] - data[1]) + abs(data[1] - data[2]) + abs(data[2] - data[0]) < 32;
 }

int isGrey(IMAGE *img)
 {
  if(img)
   {
    JSAMPLE *start = img->data, *end = img->data + 3 * img->w * img->h;
    
    for(; start != end; start += 3)
     if(!isGreyPixel(start))
      return 0;
    
    return 1;
   }
  else return 0;
 }
 
void getMinSample(JSAMPLE * pixel, IMAGE *img, double fx, double fy, double fw, double fh)
 {
  double r = 0, g = 0, b = 0;
  double samples = 0;
  int x = (int)fx, y = (int) fy;
  
  for(y = (int)fy; y <= (int)(fy+fh); ++y)
   {
    double h;
    JSAMPLE *data = img->data + (img->w * y + (int)fx)* 3;
    
    if(y == (int)fy)
     h = 1-(fy-y);
    else if(y == (int)(fy+fh))
     h = fy+fh-y;
    else
     h = 1;
    
    for(x = (int)fx; x <= (int)(fx+fw); ++x, data += 3)
     {
      double size;
      
      if(x == (int)fx)
       size = h * (1 - (fx - x));
      else if(x == (int)(fx+fw))
       size = h * (fx+fw - x);
      else
       size = h;
       
      //printf("size: %f\n", size);
      
      r += data[0]*size;
      g += data[1]*size;
      b += data[2]*size;
      samples += size;
     }
   }
  
  if(samples > 0.001)
   {
    pixel[0] = (int)(r/samples);
    pixel[1] = (int)(g/samples);
    pixel[2] = (int)(b/samples);
   }
  else pixel[0] = pixel[1] = pixel[2] = 255;
 }
 
void getMagSample(JSAMPLE *c, IMAGE *img, double fx, double fy)
 {
  int x = fx;
  int y = fy;

  JSAMPLE *a[4] = {img->data + (y * img->w + x) * 3, 
                   img->data + (y * img->w + x+1) * 3, 
                   img->data + ((y+1) * img->w + x) * 3, 
                   img->data + ((y+1) * img->w + x+1) * 3};
                   
  fx -= x;
  fy -= y;
  
  c[0] = (int)((a[0][0]*(1.0-fx)+a[1][0]*fx)*(1.0-fy)+
               (a[2][0]*(1.0-fx)+a[3][0]*fx)*fy);
         
  c[1] = (int)((a[0][1]*(1.0-fx)+a[1][1]*fx)*(1.0-fy)+
               (a[2][1]*(1.0-fx)+a[3][1]*fx)*fy);
         
  c[2] = (int)((a[0][2]*(1.0-fx)+a[1][2]*fx)*(1.0-fy)+
               (a[2][2]*(1.0-fx)+a[3][2]*fx)*fy);
 }
 
int resize_image(IMAGE *img, int w, int h)
 {
  if(img && (img->w != w || img->h != h) && w > 0 && h > 0)
   {
    JSAMPLE *newdata = malloc(w * h * 3);
    
    double region_w = (double)img->w/(double)w, region_h = (double)img->h/(double)h;
    int y = 0, x;
    
    if(!newdata) return 0;
    
    if(region_w * region_h < 1)
     {
      for(; y < w; ++y)
       for(x = 0; x < h; ++x)
        getMagSample(newdata + (y * w + x)*3, img,
                     (double)x*region_w,
                     (double)y*region_h);
     }
    else
     for(; y < w; ++y)
      for(x = 0; x < h; ++x)
       getMinSample(newdata + (y * w + x)*3, img,
                    (double)x*region_w,
                    (double)y*region_h,
                    region_w, region_h);
    
    free(img->data);
    img->data = newdata;
    img->w = w;
    img->h = h;
    return 1;
   }
  else return 0;
 }
 
IMAGE *duplicate_image(IMAGE *image)
 {
  if(image && image->data)
   {
    IMAGE *copy = malloc(sizeof(IMAGE));
    
    if(!copy)
     return 0;
    
    copy->w = image->w;
    copy->h = image->h;
    
    // allocating more mem than needed to counteract bug in resize_image. fix later.
    if(!(copy->data = malloc((image->w+1) * (image->h+1) * 3))) 
     {
      free(copy);
      return 0;
     }
     
    memcpy(copy->data, image->data, image->w * image->h * 3);
    return copy;
   }
   
  return 0;
 }
 
IMAGE *load_image(const char *filename)
 {
  IMAGE *img = malloc(sizeof(IMAGE));
  
  if(img)
   {
    FILE *infile;
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
    JSAMPROW row_pointer[1];

    if(!(infile = fopen(filename, "rb")))
     {
      free(img);
      return 0;
     }

    cinfo.err = jpeg_std_error(&jerr);
    img->w = img->h = 0;
    img->data = 0;
    
    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, infile);
    jpeg_read_header(&cinfo, TRUE);
    jpeg_start_decompress(&cinfo);
    
    img->w = cinfo.output_width;
    img->h = cinfo.output_height;
    
    if(!(img->data = malloc(img->w * img->h * 3)))
     {
      free(img);
      jpeg_destroy_decompress(&cinfo);
      fclose(infile);
      return 0;
     }

    while(cinfo.output_scanline < cinfo.output_height)
     {
      *row_pointer = img->data + 3 * img->w * cinfo.output_scanline;
      jpeg_read_scanlines(&cinfo, row_pointer, 1);
      
      if(cinfo.out_color_space == JCS_GRAYSCALE)
       ExpandRow(*row_pointer, *row_pointer + img->w);
     }

    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    fclose(infile);
     
    return img;
   }
  else return 0;
 }

int save_image(IMAGE *img, const char *filename)
 {
  if(img)
   {
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    FILE *outfile;
    JSAMPROW row_pointer[1];

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);

    if(!(outfile = fopen(filename, "wb")))
     {
      jpeg_destroy_compress(&cinfo);    
      return 0;
     }
    
    jpeg_stdio_dest(&cinfo, outfile);
    cinfo.image_width = img->w;
    cinfo.image_height = img->h;
    cinfo.in_color_space = isGrey(img) ? JCS_GRAYSCALE: JCS_RGB;
    cinfo.input_components = cinfo.in_color_space == JCS_RGB ? 3 : 1;
    
    jpeg_set_defaults(&cinfo);
    
    jpeg_set_quality(&cinfo, 99, 1);
    jpeg_simple_progression(&cinfo);
    cinfo.dct_method = JDCT_FLOAT;
    cinfo.optimize_coding = 1;
    cinfo.restart_interval = 0;
    cinfo.restart_in_rows = 0;
    
    jpeg_start_compress(&cinfo, 1);

    while(cinfo.next_scanline < cinfo.image_height)
     {
      *row_pointer = img->data + 3 * img->w * cinfo.next_scanline;
      
      if(cinfo.in_color_space == JCS_RGB)
       jpeg_write_scanlines(&cinfo, row_pointer, 1);
      else
       {
        CondenseRow(*row_pointer, *row_pointer + img->w);
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
        ExpandRow(*row_pointer, *row_pointer + img->w);
       }
     }

    jpeg_finish_compress(&cinfo);
    fclose(outfile);
    jpeg_destroy_compress(&cinfo);
    return 1;
   }
  else return 0;
 }
 
void free_image(IMAGE *img)
 {
  if(img)
   {
    if(img->data)
     free(img->data);
     
    free(img);
   }
 }
