#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include "image.h"

// This was mostly written my me, smart_idiot. The sieve code was done by Melekor.
// I don't think I used any code written by Senses777, but he wrote the original
// 2D version, and it was his idea, so I'd better mention him too.

// http://smart-idiot.no-ip.com/primecloud/

typedef unsigned __int64 bignum;

const int MEGS_TO_USE = 128;
const int WIDTH = 800;
const int HEIGHT = 600;

unsigned int *mem = 0;
unsigned int *buffer = 0;

bignum VALUE(bignum bitloc)
 {
  return (bitloc*2)+1;
 }

bignum BITLOC(bignum value)
 {
  return (value-1)/2;
 }

unsigned int *getloc(bignum bit)
 {
  return mem + (bit / (sizeof(unsigned int)*8));
 }

void clearbit(bignum bit)
 {
  *getloc(bit) &= ~(1<<(bit&(8*sizeof(unsigned int)-1)));
 }

char getbit(bignum bit)
 {
  return (*getloc(bit) & (1<<(bit&(8*sizeof(unsigned int)-1)))) != 0;
 }

void message(const char *string)
 {
  puts(string);
  fflush(stdout);
 }

int init(void)
 {
  const bignum BITSET_BYTE_COUNT = MEGS_TO_USE*1024*1024;
  const bignum BITSET_BIT_COUNT = BITSET_BYTE_COUNT*8;
  const bignum HIGHEST_BIT_TO_TEST = BITSET_BIT_COUNT-1;
  const bignum HIGHEST_VALUE_TO_TEST = (HIGHEST_BIT_TO_TEST-1)/2;
  bignum i = 1, k, sqrtval = sqrt(HIGHEST_VALUE_TO_TEST)/2;
  
  message("Generating prime numbers. . .");
    
  if(!(mem = malloc(BITSET_BYTE_COUNT)))
   {
    message("Out of memory!");
    return 0;
   }
  
  memset(mem, 0xFF, BITSET_BYTE_COUNT);

  for(; i <= sqrtval; i++)
   if(getbit(i))
    {
     bignum val = VALUE(i);
     
     for(k = val*2+val; k <= HIGHEST_VALUE_TO_TEST; k += val*2)
      clearbit(BITLOC(k));
    }
    
  if(!(buffer = malloc(WIDTH * HEIGHT * sizeof *buffer)))
   {
    message("Out of memory!");
    free(mem);
    return 0;
   }
    
  return 1;
 }
 
void uninit(void)
 {
  message("Freeing resources. . .");
  free(mem);
  free(buffer);
 }
 
void init_buffer(void)
 {
  message("Resetting buffer. . . ");
  memset(buffer, 0x00, WIDTH * HEIGHT * sizeof *buffer);
 }
 
void finish_buffer(void)
 {
  int c = 0;
  
  message("Converting buffer to image data. . . ");
  
  for(; c < WIDTH * HEIGHT; ++c)
   {
    int p = buffer[c]/32 > 255 ? 255 : buffer[c]/32;
    
    ((char *)buffer)[c*3+0] = 0;
    ((char *)buffer)[c*3+1] = p;
    ((char *)buffer)[c*3+2] = 0;
   }
 }

int x = 0, y = 0, z = 0;
float ccos = 0.5;
float csin = 0.5;

void draw_point(void)
 {
  float tx = (x * ccos - z * csin) * WIDTH * 2,
        tz = (z * ccos + x * csin) * 2 + 2000,
        ty = (y - 300) * WIDTH * 2;  

  if(tz > 0)
   {
    int _x = WIDTH/2+tx/tz, _y = HEIGHT/2+ty/tz;
      
    float rx = tx/tz-floor(tx/tz), rxi = 1-rx;
    float ry = ty/tz-floor(ty/tz), ryi = 1-ry;
	 
    if(_x >= 0 && _y >= 0 && _x < WIDTH && _y < HEIGHT)
     {
      buffer[_x+_y*WIDTH]+=WIDTH*16/tz*(rxi+ryi);
        
      if(_x+1 < WIDTH)
       {
        buffer[_x+1+_y*WIDTH]+=WIDTH*16/tz*(rx+ryi);
          
        if(_y+1 < HEIGHT)
         {
          buffer[_x+_y*WIDTH+WIDTH]+=WIDTH*16/tz*(rxi+ry);
          buffer[_x+_y*WIDTH+WIDTH+1]+=WIDTH*16/tz*(rx+ry);
         }
       }
      else if(_y+1 < HEIGHT)
       buffer[_x+_y*WIDTH+WIDTH]+=WIDTH*16/tz*(rxi+ry);
     }     
   }
 }
 
void move(char dir)
 {
  switch(dir)
   {
    case 1:
     ++x;
    break;
    
    case 2:
     --x;
    break;
        
    case 3:
     ++y;
    break;
    
    case 4:
     --y;
    break;
    
    case 5:
     ++z;
    break;
    
    case 6:
     --z;
    break;
   }
   
  draw_point();
 }

void render(void)
 {
  bignum i = 1;
  x = y = z = 0;
  init_buffer();
  message("Rendering. . . ");
  
  move(1%7);
  move(2%7);
  
  for(; i <= MEGS_TO_USE*1024*1024*8-1; ++i)
   if(getbit(i))
    move(VALUE(i)%7);
  
  finish_buffer();
 }

int save(const char *name)
 {
  IMAGE i;
  FILE *file;
  i.w = WIDTH;
  i.h = HEIGHT;
  i.data = (JSAMPLE *)buffer;
  
  if(!save_image(&i, name))
   return 0;
 
  if(!(file = fopen("list.lst", "a")))
   return 0;
         
  fprintf(file, "%s\n", name);
  fclose(file);
      
  return 1;
 }

int main(int argc, char *argv[])
 {
  const int max_frame = 800;
  int start_frame = 0, frame = 0, end_frame = max_frame;
  const double angle_inc = M_PI * 2.0 / (max_frame + 1);
  time_t start_time;
  
  if(argc == 3)
   {
    char *e = argv[1];
    start_frame = strtol(e, &e, 10);
    if(argv[1] == e)
     {
      message("Reading start frame.");
      return EXIT_FAILURE;     
     }
     
    e = argv[2];
    end_frame = 1+strtol(e, &e, 10);
    if(argv[2] == e)
     {
      message("Reading start frame.");
      return EXIT_FAILURE;     
     }
     
    if(start_frame < 0)
     {
      message("Start frame too low.");
      return EXIT_FAILURE;     
     }
     
    if(end_frame > max_frame)
     {
      message("End frame too high.");
      return EXIT_FAILURE;     
     }
     
    if(start_frame > end_frame)
     {
      message("Start frame must be less than or equal to end frame.");
      return EXIT_FAILURE;     
     }
   }
  else if(argc != 1)
   {
    message("Takes two arguments: [startframe] [endframe]");
    return EXIT_FAILURE;
   }
  else
   {
    message("No arguments given; rendering all frames.");
   }
  
  if(!init())
   {
    message("Unable to render, aborting.");
    return EXIT_FAILURE;
   }
   
  start_time = time(0);
  
  for(frame = start_frame; frame < end_frame; ++frame, angle_inc)
   {
    char name[256];
    
    if(frame != start_frame)
     {
      int elapsed = time(0) - start_time;
      int total = elapsed/((double)(frame-start_frame)/(double)(end_frame-start_frame));
      int remaining = total - elapsed;
      
      snprintf(name, sizeof name, "Preparing frame %d of %d [%d%%] [Elapsed: %0.2fh Remaining: %0.2fh Total: %0.2fh]", frame-start_frame+1, end_frame-start_frame, (frame-start_frame)*100/(end_frame-start_frame), elapsed/60.0/60.0, remaining/60.0/60.0, total/60.0/60.0);
      message(name);
     }
    else
     {
      snprintf(name, sizeof name, "Preparing frame %d of %d [%d%%]. . .", frame-start_frame+1, end_frame-start_frame, (frame-start_frame)*100/(end_frame-start_frame));
      message(name);
     }
    
    ccos = cos(frame * angle_inc);
    csin = sin(frame * angle_inc);
    render();
    
    message("Saving frame. . .");
    
    snprintf(name, sizeof name, "./frame%.4d.jpeg", frame);
    while(!save(name))
     {
      message("Error writing frame to file. Will try again in a minute (Try freeing some resources).");
      sleep(60);
     }
   }
  
  uninit();
  message("Done.");
  return EXIT_SUCCESS;
 }
